// BGG wmm
extern "C" {
	int wmmread(int, char*, int);
	int wmmlseek(int, int, int);
	int wmmopen(char*);
	int wmmgetlength(int);
	extern void post(char *fmt, ...);
}
inline unsigned uswapit(unsigned ul) {
    return (ul >> 24) | ((ul >> 8) & 0xff00) | ((ul << 8) & 0xff0000) | (ul << 24);
}

inline int iswapit(int x) { return uswapit((unsigned)x); }

inline float fswapit(float uf) {
    union { unsigned l; float f; } u;
    u.f = uf;
    u.l = uswapit(u.l);
    return (u.f);
}


// DataSet.C

#include <ugens.h>
#include "DataSet.h"
#include <stdio.h>
#include <unistd.h>		// for lseek
#include <sys/stat.h>	// for stat for getFrame()
#include <fcntl.h>

#include "lp.h"

DataSet::DataSet()
	: _nPoles(0), _frameCount(0), _fdesc(-1), _lpHeaderSize(0), 
	  _array(NULL), _oldframe(0), _endframe(0)
{
	_fprec = 22;
}

DataSet::~DataSet()
{
	if (_fdesc > 0)
		::close(_fdesc);
	delete [] _array;
}

int
DataSet::open(const char *fileName, int npoleGuess, float sampRate)
{
// BGG wmm
//    if ((_fdesc = ::open(fileName, O_RDONLY)) < 0) {
    if ((_fdesc = wmmopen((char *)fileName)) < 0) {
		::rterror("dataset", "Can't open %s", fileName);
		return -1;
    }
	_nPoles = npoleGuess;	// in case we are not using headers
	::rtcmix_advise("dataset", "Opened lpc dataset %s.", fileName);
#ifdef USE_HEADERS
	if ((_lpHeaderSize = ::checkForHeader(_fdesc, &_nPoles, sampRate)) < 0) {
	    ::rterror("dataset", "Failed to check header");
		return -1;
	}
#else
	if (!_nPoles) {
		return -1;
	}
#endif /* USE_HEADERS */

	allocArray(_nPoles);

	_npolem1=_nPoles-1;
	_framsize=_nPoles+4;
	_recsize=_fprec*_framsize;
	_bprec=_recsize*FLOAT;
	_bpframe=_framsize*FLOAT;

	struct stat st;
	/* store and return number of frames in datafile */
// BGG wmm
/*
	if (::stat(fileName, &st) >= 0) {
		_frameCount = (st.st_size-_lpHeaderSize) / _bpframe;
		return _frameCount;
	}
*/
	_frameCount = (wmmgetlength(_fdesc) - _lpHeaderSize) / _bpframe;
	if (_frameCount > 0) return _frameCount;

	else {
		::rterror("dataset", "Unable to stat dataset file.");
		return -1;
	}
}

int
DataSet::getFrame(float frameno, float *pCoeffs)
{
	int i,j;
	int frame = (int)frameno;
	float fraction = frameno - (float)frame;
	if (!((frame >= _oldframe) && (frame < _endframe))) {
		int bytesRead, framesRead = _fprec;
// BGG wmm
//    	if (::lseek(_fdesc, _lpHeaderSize+(frame*_bpframe), 0) == -1)
    	if (wmmlseek(_fdesc, _lpHeaderSize+(frame*_bpframe), 0) == -1)
		{
// BGG wmm -- changed the fprintf() to post()
            	post("bad lseek on LPC analysis file \n");
            	return(-1);
    	}
#ifdef debug
		printf("Reading frames %d - %d from file\n", frame, frame + 22);
#endif
		/* Quit if we read less than one complete frame */
// BGG wmm
//    	if ((bytesRead = ::read(_fdesc, _array, _bprec)) < _bpframe)
    	if ((bytesRead = wmmread(_fdesc, (char *)_array, _bprec)) < _bpframe)
		{
// BGG wmm -- changed the fprintf() to post()
            	post("reached eof on LPC analysis file \n");
            	return(-1);
    	}

// BGG wmm -- swap 'em
	for (i = 0; i < _recsize; i++) _array[i] = fswapit(_array[i]);

		framesRead = bytesRead / _bpframe;
    	_oldframe = frame;
    	_endframe = _oldframe + framesRead - 1;
	}
	for(i=(frame-_oldframe)*_framsize,j=0; j<_framsize; i++,j++)  
        	pCoeffs[j] = _array[i] + fraction * (_array[i+_framsize] - _array[i]);
	return(0);
}

void
DataSet::allocArray(int nPoles)
{
	if (_array)
	    delete [] _array;
	_array = new float[_fprec * (nPoles + 4)];
}
