/*   See ``AUTHORS'' for a list of contributors. See ``LICENSE'' for
   the license to this software and for a DISCLAIMER OF ALL WARRANTIES.
*/
//#define DBUG
//#define DENORMAL_CHECK
// BGG wmm -mno-cygwin -- no pthreads
//#include <pthread.h>
// BGG wmm -mno-cygwin -- use local version
//#include <sys/resource.h>
#include <wmm_sys_resource.h>
#include <ctype.h>
#include <string.h>
#include <math.h>
#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
// BGG wmm -mno-cygwin -- no sockets or threads
//#include <sys/socket.h>
//#include <netinet/in.h>
#include <iostream.h>
#include <signal.h>

#include <RTcmixMain.h>
#include <prototypes.h>
#include <ugens.h>
#include <ug_intro.h>
#include <version.h>
#include "rt.h"
#include "heap.h"
#include "sockdefs.h"
#include "notetags.h"           // contains defs for note-tagging
#include "dbug.h"


extern "C" {
#ifdef SGI
  void flush_all_underflows_to_zero();
#endif
#ifdef LINUX
  void sigfpe_handler(int sig);
#endif
}


/* ----------------------------------------------------- detect_denormals --- */
/* Unmask "denormalized operand" bit of the x86 FPU control word, so that
   any operations with denormalized numbers will raise a SIGFPE signal,
   and our handler will be called.  NOTE: This is for debugging only!
   This will not tell you how many denormal ops there are, so just because
   the exception is thrown doesn't mean there's a serious problem.  For
   more info, see: http://www.smartelectronix.com/musicdsp/text/other001.txt.
*/
#ifdef LINUX
#ifdef DENORMAL_CHECK
static void
detect_denormals()
{
   #include <fpu_control.h>
   int cw = 0;
   _FPU_GETCW(cw);
   cw &= ~_FPU_MASK_DM;
   _FPU_SETCW(cw);
}
#endif /* DENORMAL_CHECK */
#endif /* LINUX */


/* ----------------------------------------------------------------- main --- */

// BGG wmm
#define TOTOBJS 178 // from rtcmix~.c. max/msp
RTcmixMain *app[TOTOBJS];
// BGG wmm
#include <wmmcontext.h>
wmmcontext theContext[TOTOBJS];


extern "C" {

void post(char *fmt, ...);

// BGG wmm -- the "objno" arg for this and subsequent functions enable
//		the 'context-switching' for the windows version
int
rtcmixmain(int objno)	// BGG mm -- now called this for max/msp
{
#ifdef LINUX
 #ifdef DENORMAL_CHECK
   detect_denormals();
 #endif
   signal(SIGFPE, sigfpe_handler);          /* Install signal handler */
#endif /* LINUX */
#ifdef SGI
   flush_all_underflows_to_zero();
#endif


// BGG no argc and argv in max/msp version mm
   app[objno]->clear_wmmcontext(objno);
   app[objno] = new RTcmixMain();
   app[objno]->run(); // in max/msp this just sets it all up...

// BGG wmm the context stuff
   app[objno]->set_wmmcontext(objno);

   return 0;
}


// BGG -- these will be called from max/msp, an instrument will set
// the *_ready value to signal to return something other than NULL
// max/msp will then respond accordingly

// BGG wmm -- bang_ready is now in RTcmix.cpp
//int bang_ready = 0;
extern int bang_ready;

int check_bang(int objno)
{  
   app[objno]->get_wmmcontext(objno);

   if (bang_ready == 1) {
      bang_ready = 0;
      app[objno]->set_wmmcontext(objno);
      return(1);
   } else {
// BGG wmm
      app[objno]->set_wmmcontext(objno);
      return(0);
   }
}


// BGG wmm -- these are now in RTcmix.cpp
//int vals_ready = 0;
//float maxmsp_vals[MAXDISPARGS];
extern int vals_ready;
extern float *maxmsp_vals;

int check_vals(int objno, float thevals[])
{  
   int i, tvals;
   
   app[objno]->get_wmmcontext(objno);
   if (vals_ready > 0) { // vals_ready will contain how many vals to return
      for (i = 0; i < vals_ready; i++) thevals[i] = maxmsp_vals[i];
      tvals = vals_ready;
      vals_ready = 0;
      app[objno]->set_wmmcontext(objno);
      return(tvals);
   } else {
// BGG wmm
      app[objno]->set_wmmcontext(objno);
      return(0);
   }
}


int err_pending = 0;

int check_error()
{
// err_pending is set in sys/message.c, the max/msp buf theerror[] contains
// the text   if terr == 1, do a warn, if terr == 2, full-blown error
   int terr;

   if (err_pending > 0) {
      terr = err_pending;
      err_pending = 0;
      return(terr);
   } else {
      return(0);
   }
}


float *maxmsp_outbuf; // written to in rtsendsamps (write_to_audio_device)
float *maxmsp_inbuf; // read fromin rtgetsamps

void pullTraverse(int objno, float *out, float *in)
{
	app[objno]->get_wmmcontext(objno);
	maxmsp_outbuf = out;
	maxmsp_inbuf = in;
	app[objno]->inTraverse(NULL, NULL);
	app[objno]->set_wmmcontext(objno);
}

int maxmsp_rtsetparams(float sr, int nchans, int vecsize, char *mm_errbuf, int objno)
{
	int status;

	app[objno]->get_wmmcontext(objno);

	status = (int)app[objno]->mm_rtsetparams(sr, nchans, vecsize, mm_errbuf);

	app[objno]->set_wmmcontext(objno);

	return(status);
}


// these are set from inlets on the rtcmix~ object, using PFields to
// control the Instruments
// BGG wmm -- now set and alloc'd in RTcmix.cpp
//float inletvals[10];
extern float *inletvals;

void pfield_set(int inlet, float pval, int objno)
{
	app[objno]->get_wmmcontext(objno);
	inletvals[inlet-1] = pval;
	app[objno]->set_wmmcontext(objno);
}

// BGG wmm -- may not need this
void rtcmix_closeit(int objno)
{
	app[objno]->get_wmmcontext(objno);
	delete app[objno];
	app[objno]->set_wmmcontext(objno);
}

// BGG wmm
double maxmsp_parse_dispatch(char *cmd, double *p, int nargs, void **inst, int objno)
{
	double retval;

	app[objno]->get_wmmcontext(objno);
	retval  = dispatch(cmd, p, nargs, inst);
	app[objno]->set_wmmcontext(objno);

	return retval;
}

// BGG wmm -- from parser/parse_with_minc
extern int parse_score(char *, int);

int maxmsp_parse_score(char *buf, int len, int objno)
{
	int retval;

	app[objno]->get_wmmcontext(objno);
	retval = parse_score(buf, len);
	app[objno]->set_wmmcontext(objno);

	return retval;
}

// use this to pass in info from the max/msp [buffer~] object.  All
// we need to know is the number of frames, nchans, and starting address
// flags and globals to support this stuff, arg!
// BGG wmm -- these now set in RTcmix.cpp
#define MAX_MM_BUFS 50
//mm_buf mm_bufs[MAX_MM_BUFS]; // mm_buf defined in rtdefs.h
//int mm_buf_input = -1; // used in rtgetin() and then rtsetinput()
//int n_mm_bufs = 0;
extern mm_buf *mm_bufs;
extern int mm_buf_input;
extern int n_mm_bufs;

void buffer_set(char *bufname, float *bufstart, int nframes, int nchans, int modtime, int objno)
{
	int i;
	int foundit;

	if (strlen(bufname) >= MAX_MM_BUFNAME) { // defined in rtdefs.h
		warn("bufset", "[buffer~] name has to be < %d chars", MAX_MM_BUFNAME);
		return;
	}

	// check to see if this is already set
	foundit = 0;
	app[objno]->get_wmmcontext(objno);
	for (i = 0; i < n_mm_bufs; i++) {
			if (strcmp(bufname, mm_bufs[i].name) == 0) {
				foundit = 1;
				if (modtime > mm_bufs[i].mm_modtime) { // buffer was modified
					mm_bufs[i].mm_bufstart = bufstart;
					mm_bufs[i].mm_buf_nframes = nframes;
					mm_bufs[i].mm_buf_chans = nchans;
					mm_bufs[i].mm_modtime = modtime;
				}
				break;
			}
	}

	// it's a new one
	if (foundit == 0) {
		if (n_mm_bufs >= MAX_MM_BUFS) {
			warn("bufset", "we can only do %d [buffer~] buffers at present, sorry!", MAX_MM_BUFS);
			return;
		}
		strcpy(mm_bufs[n_mm_bufs].name, bufname);
		mm_bufs[n_mm_bufs].mm_bufstart = bufstart;
		mm_bufs[n_mm_bufs].mm_buf_nframes = nframes;
		mm_bufs[n_mm_bufs].mm_buf_chans = nchans;
		mm_bufs[n_mm_bufs].mm_modtime = modtime;
		n_mm_bufs++;
	}
	app[objno]->set_wmmcontext(objno);
}

// called for the [flush] message; deletes and reinstantiates the rtQueue
// and rtHeap, thus flushing all scheduled events in the future
void flush_sched(int objno)
{
	app[objno]->get_wmmcontext(objno);
	app[objno]->resetQueueHeap(); // in RTcmixMain.cpp
	app[objno]->set_wmmcontext(objno);
}


} // end of "extern C"  BGG mm
